function [H, gcells, gfaces, gnodes] = extractSubgrid(G, c)
%Construct valid grid definition from subset of existing grid cells.
%
% SYNOPSIS:
%    subG              = extractSubgrid(G, cells)
%   [subG, gc, gf, gn] = extractSubgrid(G, cells)
%
% PARAMETERS:
%   G     - Valid grid definition.
%
%   cells - Subset of existing grid cells for which the subgrid data
%           structure should be constructed.  Must be an array of numeric
%           cell indices.  A logical mask into the range 1:G.cells.num is
%           not supported.  Moreover, the cell indices must be unique
%           within the subset.  Repeated indices are not supported.
%
% RETURNS:
%   subG  - Resulting subgrid.  The cells of 'subG' are ordered according
%           to increasing numeric index in 'cells'.
%
%   gcells, 
%   gfaces,
%   gnodes -
%           Global cells, faces, and nodes referenced by sub-grid 'subG'.
%           Specifically,
%
%              gcells(i) == global cell index (from 'G') of subG cell 'i'.
%              gfaces(i) == global face index (from 'G') of subG face 'i'.
%              gnodes(i) == global node index (from 'G') of subG node 'i'.
%
% OPTIONAL:
%
%          none.
%
% NOTE.
%   The return value 'gcells' is equal to the input parameter 'cells'.
%
% EXAMPLE:
%   G    = cartGrid([3, 5, 7]);     [I, J, K] = ndgrid(2, 2:4, 3:5);
%   subG = extractSubgrid(G, sub2ind(G.cartDims, I(:), J(:), K(:)));
%   plotGrid(subG, 'EdgeAlpha', 0.1, 'FaceAlpha', 0.25)
%   view(-35,20), camlight
%
% SEE ALSO:
%   grid_structure.

%{
Copyright 2009, 2010, 2011 SINTEF ICT, Applied Mathematics.

This file is part of The MATLAB Reservoir Simulation Toolbox (MRST).

MRST is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

MRST is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with MRST.  If not, see <http://www.gnu.org/licenses/>.
%}

% $Date: 2011-12-31 17:43:06 +0100 (Sat, 31 Dec 2011) $
% $Revision: 8727 $

   % We don't support multi-component grids...
   if numel(G) > 1,
      error(msgid('Grid:MultiComponent'), ...
            'Cannot extract sub-grid from more than one grid at a time.');
   end

   if ~isfield(G.cells, 'facePos'),
      error(msgid('Grid:Cells:NoFacePos'), ...
           ['Grid must support ''facePos'' field in order to ', ...
            'extract sub-grid.']);
   end

   if ~isfield(G.faces, 'nodePos'),
      error(msgid('Grid:Faces:NoNodePos'), ...
           ['Grid must support ''nodePos'' field in order to ', ...
            'extract sub-grid.']);
   end

   ix = @(p,i) mcolon(double(p(i)), double(p(i+1)) - 1);

   % Sort the subset cells, c, in ascending order.  Add one for outside.
   cells      = false([G.cells.num + 1, 1]);
   cells(c+1) = true;

   % Extract faces connected to 'c'.
   faces = false([G.faces.num, 1]);
   faces(G.cells.faces(ix(G.cells.facePos, c))) = true;

   % Extract nodes connected to 'faces' (in cell subset).
   nodes = false([G.nodes.num, 1]);
   ff    = find(faces);
   nodes(G.faces.nodes(ix(G.faces.nodePos, ff))) = true;

   H.nodes.coords    = G.nodes.coords(nodes, :);
   if isfield(G.cells, 'numFaces'),
      warning('MRST:deprecated','The field numFaces will not be supported in future releases of MRST.')
      numFaces = G.cells.numFaces(cells(2:end));
   else
      numFaces = diff(G.cells.facePos);
      numFaces = numFaces(cells(2:end));
   end

   if isfield(G.faces, 'numNodes'),
      numNodes = G.faces.numNodes(faces);
   else
      numNodes = diff(G.faces.nodePos);
      numNodes = numNodes(faces);
   end
   H.faces.neighbors = G.faces.neighbors(faces, :);

   pos = @(n) int32(cumsum([1; double(reshape(n, [], 1))]));
   H.cells.facePos   = pos(numFaces);
   H.faces.nodePos   = pos(numNodes);

   H.cells.faces       = G.cells.faces(ix(G.cells.facePos, c ), :);
   H.faces.nodes       = G.faces.nodes(ix(G.faces.nodePos, ff)   );

   % Renumbering of grid entities (cells,faces,nodes) in subgrid.
   mc = zeros([G.cells.num+1, 1], 'int32');   mc(cells) = 1:numel(c);
   fc = zeros([G.faces.num  , 1], 'int32');   fc(faces) = 1:numel(ff);
   nc = zeros([G.nodes.num  , 1], 'int32');   nc(nodes) = 1:sum(nodes);

   H.faces.neighbors  = mc(H.faces.neighbors+1);
   H.cells.faces(:,1) = fc(H.cells.faces(:,1));
   H.faces.nodes      = nc(H.faces.nodes);

   H.faces.num = numel(H.faces.nodePos) - 1;
   H.cells.num = numel(H.cells.facePos) - 1;
   H.nodes.num = size(H.nodes.coords, 1);

   try
      H.cells.indexMap = G.cells.indexMap(cells(2:end));
      H.cartDims = nan(size(G.cartDims));
   end
   
   % Record history.
   assert(isfield(G, 'type'), ...
      'Every grid must record its origin in the field ''type''.');
   H.type    = [G.type, mfilename];
   
   assert(isfield(G, 'griddim'), ...
      'Every grid must declare its intrinsic dimension in the field ''griddim''.');
   H.griddim = G.griddim;
   
   % Preserve 'computeGeometry' fields, if present.
   if isfield(G.cells, 'volumes'),
      H.cells.volumes   = G.cells.volumes  (cells(2:end)  );
   end
   if isfield(G.cells, 'centroids'),
      H.cells.centroids = G.cells.centroids(cells(2:end),:);
   end
   if isfield(G.faces, 'areas'),
      H.faces.areas     = G.faces.areas    (faces  );
   end
   if isfield(G.faces, 'tag'),
      H.faces.tag     = G.faces.tag    (faces  );
   end
   if isfield(G.faces, 'normals'),
      H.faces.normals   = G.faces.normals  (faces,:);
   end
   if isfield(G.faces, 'centroids'),
      H.faces.centroids = G.faces.centroids(faces,:);
   end

   gcells = find(mc) - 1;
   gfaces = find(fc);
   gnodes = find(nc);
end
