function [G, W, rock] = setupSPE10layers(varargin)
% setupSPE10layer -- Initialise properties for Model 2 of Tenth SPE
%            Comparative Solution Project.
%
% SYNOPSIS:
%   [G, W, rock] = setupSPE10layers
%   [G, W, rock] = setupSPE10layers(layers)
%
% PARAMETERS:
%   layers - Which of the 85 model layers to include in a specific test.
%            OPTIONAL.  If unspecified, all 85 layers (a total of
%            60-by-220-by-85 == 1,122,000 cells) are included.
%
%            Some possible values are
%               layers = ( 1 : 35).';  %  Tarbert formation
%               layers = (36 : 85).';  %  Upper Ness formation
%
% RETURNS:
%   G    - grid structure as described in grid_structure.
%   W    - Well structure.  Injector at 500 Bar, producers at 200 Bar.
%   rock - Rock structure having fields 'perm' and 'poros' pertaining to
%          the specified layers.
%
% EXAMPLE:
%   [G, W, rock] = setupSPE10layers(85)
%
% SEE ALSO:
%   SPE10_rock, computeGeometry.

% $Date: 2011-12-20 20:52:40 +0100 (Tue, 20 Dec 2011) $
% $Revision: 8651 $

mrstModule add spe10
%% Define layers
if nargin > 0 && isnumeric(varargin{1}),
   layers = varargin{1};
else
   layers = (1 : 85) .';
end

%% Define grid dimensions and well locations
cartDims = [  60,  220,   numel(layers)];
physDims = [1200, 2200, 2*cartDims(end)] .* ft();   % ft -> m

wtype    = {'bhp', 'bhp', 'bhp', 'bhp', 'bhp'};
wtarget  = [200,   200,   200,   200,   500] .* barsa();
wrad     = [0.125, 0.125, 0.125, 0.125, 0.125] .* meter;
wloc     = [  1,   60,     1,   60,   30;
              1,    1,   220,  220,  110];
wname    = {'P1', 'P2', 'P3', 'P4', 'I1'};

%% Get reservoir rock properties in given layers
rock      = SPE10_rock(layers);
rock.perm = convertFrom(rock.perm, milli*darcy);

%% Define grid
G = cartGrid(cartDims, physDims);
G = computeGeometry(G);

%% Define wells
% All wells completed in all layers of the model

W = [];
for w = 1 : numel(wtype),
   W = verticalWell(W, G, rock, wloc(1,w), wloc(2,w), 1 : numel(layers), ...
                    'Type', wtype{w}, 'Val', wtarget(w), ...
                    'Radius', wrad(w), 'Name', wname{w});
end
