function perm = upscalePermSB(g, cg, rock, varargin)
% Compute upscaled Stokes-Brinkman permeabilites using flow based upscaling.
%
% SYNOPSIS:
%   perm_coarse = upscalePermSB(G, CG, rock, fluid)
%   perm_coarse = upscalePermSB(G, CG, rock, fluid, 'pn1', pv1, ...)
%
% PARAMETERS:
%   G, CG    - Underlying grid (G, see 'grid_structure') and coarse grid
%              model (CG) defined by function 'generateCoarseGrid'.
%
%   rock     - Rock data structure with valid field 'perm'.  The
%              permeability is assumed to be in measured in units of
%              metres squared (m^2).
%
%   'pn'/pv  - List of 'key'/value pairs defining optional parameters. 
%
%                - Verbose -- 
%                         Whether or not to emit progress reports during
%                         the assembly process.
%                         Logical.  Default value dependent upon global
%                         verbose settings of function 'mrstVerbose'.
%
% RETURNS:
%   perm_coarse - Upscaled permeability field
%
% REMARK:
%   The flow based upscaling applied here is based on the assumption that
%   the grid is Cartesian. 
%
% SEE ALSO:
%   upscalePerm, extractSubgrid, computeMimeticIP, mrstVerbose.

%{
Copyright 2009, 2010, 2011 SINTEF ICT, Applied Mathematics.

This file is part of The MATLAB Reservoir Simulation Toolbox (MRST).

MRST is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

MRST is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with MRST.  If not, see <http://www.gnu.org/licenses/>.
%}

% $Date: 2011-12-31 17:43:06 +0100 (Sat, 31 Dec 2011) $
% $Revision: 8727 $

   assert ( isfield(g, {'cartDims'}))

   opt = struct('Verbose', mrstVerbose);   
   opt = merge_options(opt, varargin{:});
   
   if opt.Verbose,
      h = waitbar(0, 'Computing upscaled Stokes-Brinkman permeabilities...');
      fprintf('Computing upscaled Stokes-Brinkman permeabilities... '), tic
   end

   fluid    = initSingleFluid('mu' , 1 * Pascal*second, ...
                              'rho', 1 * kilogram/meter^3);
   
   dims     = find(g.cartDims);
   nBlocks  = cg.cells.num; 
   perm     = zeros(cg.cells.num, numel(g.cartDims));
   
   ups = @(cells, K_cells) upscale_block(extractSubgrid(g, cells), ...
                                         K_cells, dims, fluid);
   map.sub_cells = @(b) find(cg.partition == b);
  
   
   % Compute upscaled permeability for each coarse block
   for b = 1 : nBlocks,
            
      cells = map.sub_cells(b);  % Fine-scale cells present in 'blk'
      K_cells = rock.perm(cells, :);
          
      perm(b,:) = ups(cells, K_cells); 
            
      if opt.Verbose, waitbar(b / nBlocks, h), end
   end
   if opt.Verbose,
      toc, close(h)
     
   end
end


%--------------------------------------------------------------------------
% Helpers follow.
%--------------------------------------------------------------------------

function k = upscale_block(subG, K_cells, dims, fluid)
%  
%  Compute perm in x-direction:
%
%                  v*n = 0    
%              ______________  
% l_boundary  |              | r_boundary          
%             |              |
%     p1 = 1  |              | p2 = 0
%             |______________|
%                   v*n = 0
% 
%             |--------------|    
%                    L  
%                    
%           grad p = (p2 - p1)/L 
%           q1: flux over right boundary
%           kx* = (q1*L)/(area(r_boundary)*(p2 - p1))
%     
fluid.mu_eff = fluid.properties(initResSol(subG, 0.0));  % HACK ALERT !!!
Dofs         = findCartDofs(subG);
rock.perm    = K_cells;
k            = zeros(1, numel(dims));

%{'LEFT', 'RIGHT', 'FRONT', 'BACK', 'BOTTOM', 'TOP'};
tags = [ 1 2; 3 4; 5 6]; 

bnd_f = any(subG.faces.neighbors==0,2);
ind   = bnd_f(subG.cells.faces(:,1));
faceAndTag = subG.cells.faces(ind, :);

for i = dims

   faces1 = faceAndTag(faceAndTag(:,2) == tags(i,1));
   faces2 = faceAndTag(faceAndTag(:,2) == tags(i,2));
   ind        = bnd_f;
   ind(faces1) = false;
   ind(faces2) = false;
   noflow_faces = find(ind);
   
   bc = addBCSB([], faces1, 'pressure', repmat(1*barsa, numel(faces1),  1), ...
                subG, Dofs);
   bc = addBCSB(bc, faces2, 'pressure', repmat(0, numel(faces2),  1), ...
                subG, Dofs);
   bc = addBCSB(bc, noflow_faces, 'velocity_n',...
                repmat(0, numel(noflow_faces),1), subG, Dofs);
   
   Ssb = makeSystemSB(subG, Dofs, rock, fluid, 'bc', bc);
   
   [Ssb, rSol] = solveSystemSB(Ssb, subG, Dofs, 'bc', bc);
   rSol        = nodeToCellData(rSol, subG, Dofs);
      
   area = sum(subG.faces.areas(faces2,:));
   L    = abs(subG.faces.centroids(faces1(1),i)...
           -subG.faces.centroids(faces2(1),i));

   q    = abs(sum(rSol.flux(faces2)));  
   k(i) = q*L/(1*barsa*area);   
end
end

%--------------------------------------------------------------------------
