function varargout = writeGRDECL(grdecl, filename)
%Write a GRDECL structure out to permanent file on disk.
%
% SYNOPSIS:
%   writeGRDECL(grdecl, file)
%
% PARAMETERS:
%   grdecl - A corner-point GRDECL structure which may be passed to
%            function 'processGRDECL' in order to construct a grid
%            structure.
%
%   file   - Name of output file.  String.  This name is passed directly to
%            function 'fopen' using mode 'wt'.
%
% RETURNS:
%   Nothing.
%
% SEE ALSO:
%   fopen, readGRDECL, processGRDECL.

%{
Copyright 2009, 2010, 2011 SINTEF ICT, Applied Mathematics.

This file is part of The MATLAB Reservoir Simulation Toolbox (MRST).

MRST is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

MRST is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with MRST.  If not, see <http://www.gnu.org/licenses/>.
%}

% $Date: 2011-12-31 17:43:06 +0100 (Sat, 31 Dec 2011) $
% $Revision: 8727 $

   success = true;

   format = @(s,n) [repmat([s, ' '], [1, n]), '\n'];

   [fid, msg] = fopen(filename, 'wt');
   if fid < 0, error(msg); end

   fprintf(fid, 'SPECGRID\n%d %d %d 1 F\n/\n\n', grdecl.cartDims);
   fprintf(fid, 'COORD\n');
   fprintf(fid, format('%12f', 6), grdecl.COORD);
   fprintf(fid, '/\n\n');

   fprintf(fid, 'ZCORN\n');
   fprintf(fid, format('%12f', 8), grdecl.ZCORN);
   fprintf(fid, '/\n\n');

   fprintf(fid, 'ACTNUM\n');
   if isfield(grdecl, 'ACTNUM'),
      actnum = grdecl.ACTNUM;
   else
      % No field 'ACTNUM', assume all cells active.
      actnum = ones([prod(grdecl.cartDims(:)), 1]);
   end
   fprintf(fid, format('%d', 20), actnum);
   fprintf(fid, '/');

   fprintf(fid, '\n\nSATNUM\n');
   if isfield(grdecl, 'SATNUM'),
      satnum = grdecl.SATNUM;
   else
      % No field 'SATNUM', assume one rock type.
      satnum = ones([prod(grdecl.cartDims(:)), 1]);
   end
   fprintf(fid, format('%d', 20), satnum);
   fprintf(fid, '/');

   if isfield(grdecl, 'PORO'),
      fprintf(fid, '\n\nPORO\n');
      fprintf(fid, format('%18.5e', 4), grdecl.PORO);
      fprintf(fid, '/');
   end

   if isfield(grdecl, 'PERMX'),
      fprintf(fid, '\n\nPERMX\n');
      fprintf(fid, format('%18.5e', 4), grdecl.PERMX);
      fprintf(fid, '/');
   end

   if isfield(grdecl, 'PERMY'),
      fprintf(fid, '\n\nPERMY\n');
      fprintf(fid, format('%18.5e', 4), grdecl.PERMY);
      fprintf(fid, '/');
   end

   if isfield(grdecl, 'PERMZ'),
      fprintf(fid, '\n\nPERMZ\n');
      fprintf(fid, format('%18.5e', 4), grdecl.PERMZ);
      fprintf(fid, '/');
   end

   fprintf(fid, '\n');
   fclose(fid);

   if nargout > 0,
      varargout{1} = success;
   end
end
