#include <string.h>
#include <mex.h>
#include <lapack.h>

/* Invert a many square matrices using LAPACK routines DGETRF and DGETRI. 
 * The matrix entries and dimensions of each matrix is input, and the matic 
 * entries of the inverses are output.
 */
void mexFunction( int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[] )
{
  if ((nrhs != 2) || (nlhs != 1))
    {
      fprintf(stderr, "return early\n");
      return;
    }
  int i,j,k,p;

  /* Get input */
  if (mxIsSparse(prhs[0]))
  {
     mexErrMsgTxt("First argument must be a (full) vector.");  
  }
  double  *A  = mxGetPr(prhs[0]);
  if(!mxIsInt32(prhs[1]))
    {
      mexErrMsgTxt("Second argument (vector of sizes) must be int32.");
    }
  int     *sz = (int*)mxGetData(prhs[1]);
  int      n  = mxGetNumberOfElements(prhs[1]);
  int     nnz = mxGetNumberOfElements(prhs[0]);

  
  /* Create output vector */
  plhs[0]   = mxCreateDoubleMatrix(nnz, 1, mxREAL);
  double *B = mxGetPr(plhs[0]);
  memcpy(B, A, nnz*sizeof(double));

  /* Invert all matrices. */
  invertFullMatrices(n, B, sz);
}



int invertFullMatrices(int n, double *a, int *sz)
{

  int i;

  /* Find maximal system size */
  int maxsize = 0;
  for (i=0; i<n; ++i)
    {
      maxsize = (sz[i] > maxsize ? sz[i] : maxsize);
    }
  size_t  LWORK = maxsize;
  size_t *IPIV  = mxMalloc(maxsize * sizeof(*IPIV)); 
  double *WORK  = mxMalloc(LWORK  * sizeof(*WORK));
  size_t  INFO;
 
  int j;
  int issue_warning = 0;
  int offset  = 0;
  for (i=0; i<n; ++i)
  {
    size_t  N    = sz[i];    
    
    /*  Factor matrix */
    dgetrf_(&N, &N, a+offset, &N, IPIV, &INFO);
    if (INFO){
       /* fprintf(stderr, "GDETRF returned nonzero INFO = %d\n", INFO); */
       /* fprintf(stderr, "N = %d\n", N); */
       for (j=0; j<sz[i]*sz[i]; ++j)   a[offset+j] = mxGetInf();
       issue_warning = 1;
    }
    
    
    else {
       /* Invert matrix */
       dgetri(&N, a+offset, &N, IPIV, WORK, &LWORK, &INFO);
       if (INFO) {
          /* fprintf(stderr, "DGETRI returned nonzero INFO = %d\n", INFO); */
          /* fprintf(stderr, "N = %d\n", N); */
          for (j=0; j<sz[i]*sz[i]; ++j)   a[offset+j] = mxGetInf();
          issue_warning = 1;
       }
    }
    offset += sz[i]*sz[i];
  }

  if (issue_warning){
     mexWarnMsgIdAndTxt("InvertSparseBlockMatrix:Singular", "Matrix block is singular to waking precision.");
  }
  
  mxFree(WORK);
  mxFree(IPIV);

  return INFO;
}
